#include <msp/core/maputils.h>
#include <msp/core/raii.h>
#include "beamgate.h"
#include "blockallocator.h"
#include "block.h"
#include "catalogue.h"
#include "driver.h"
#include "layout.h"
#include "trackcircuit.h"
#include "trackiter.h"
#include "train.h"
#include "vehicle.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

struct BlockAllocator::BlockMatch
{
	const Block &block;

	BlockMatch(const Block &b): block(b) { }

	bool operator()(const BlockIter &bi) const { return &*bi==&block; }
};


BlockAllocator::BlockAllocator(Train &t):
	train(t),
	active(false),
	cur_blocks_end(blocks.end()),
	next_sensor(0),
	pending_block(0),
	stop_at_block(0),
	reserving(false),
	advancing(false)
{
	Layout &layout = train.get_layout();
	layout.signal_block_reserved.connect(sigc::mem_fun(this, &BlockAllocator::block_reserved));
	layout.signal_sensor_state_changed.connect(sigc::mem_fun(this, &BlockAllocator::sensor_state_changed));
	layout.get_driver().signal_halt.connect(sigc::mem_fun(this, &BlockAllocator::halt_event));

	const set<Track *> &tracks = layout.get_all<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		if((*i)->get_type().is_turnout())
		{
			(*i)->signal_path_changing.connect(sigc::hide(sigc::bind(sigc::mem_fun(this, &BlockAllocator::turnout_path_changing), sigc::ref(**i))));
			(*i)->signal_path_changed.connect(sigc::hide(sigc::bind(sigc::mem_fun(this, &BlockAllocator::turnout_path_changed), sigc::ref(**i))));
		}
}

void BlockAllocator::set_active(bool a)
{
	active = a;
	if(active)
		reserve_more();
	else
	{
		release_blocks_end(cur_blocks_end);
		pending_block = 0;
	}
}

bool BlockAllocator::start_from(const BlockIter &block)
{
	if(!block)
		throw invalid_argument("BlockAllocator::start_from");

	float remaining_length = 0;
	unsigned n_vehs = train.get_n_vehicles();
	for(unsigned i=0; i<n_vehs; ++i)
		remaining_length += train.get_vehicle(i).get_type().get_length();

	clear();

	BlockList blocks_to_reserve;
	for(BlockIter b=block; b; b=b.next())
	{
		if(b->get_train())
			break;
		blocks_to_reserve.push_back(b);
		for(TrackIter t=b.track_iter(); (t && &t->get_block()==&*b); t=t.next())
			remaining_length -= t->get_path_length();
		if(remaining_length<=0)
			break;
	}

	if(remaining_length>0)
		return false;

	for(BlockList::iterator i=blocks_to_reserve.begin(); i!=blocks_to_reserve.end(); ++i)
	{
		blocks.push_back(*i);
		try
		{
			(*i)->reserve(&train);
		}
		catch(...)
		{
			blocks.pop_back();
			while(i!=blocks_to_reserve.begin())
			{
				blocks.pop_back();
				(*--i)->reserve(0);
			}
			throw;
		}
	}

	return true;
}

void BlockAllocator::rewind_to(const Block &block)
{
	if(!active)
		return;

	BlockList::iterator i = find_if(cur_blocks_end, blocks.end(), BlockMatch(block));
	if(i!=blocks.end())
	{
		release_blocks_end(i);
		reserve_more();
	}
}

void BlockAllocator::clear()
{
	release_blocks_begin(blocks.end());
	active = false;
	next_sensor = 0;
	pending_block = 0;
	stop_at_block = 0;
}

void BlockAllocator::stop_at(const Block *block)
{
	stop_at_block = block;
	if(active && !block)
		reserve_more();
}

const BlockIter &BlockAllocator::first() const
{
	if(blocks.empty())
		throw logic_error("no blocks");
	return blocks.front();
}

const BlockIter &BlockAllocator::last() const
{
	if(blocks.empty())
		throw logic_error("no blocks");
	BlockList::const_iterator i = --blocks.end();
	if(i->block()==pending_block)
		--i;
	return *i;
}

const BlockIter &BlockAllocator::last_current() const
{
	if(blocks.empty())
		throw logic_error("no blocks");
	if(cur_blocks_end==blocks.begin())
		throw logic_error("internal error (no current blocks)");
	BlockList::const_iterator i = cur_blocks_end;
	return *--i;
}

const BlockIter &BlockAllocator::iter_for(const Block &block) const
{
	BlockList::const_iterator i = find_if(blocks.begin(), blocks.end(), BlockMatch(block));
	if(i==blocks.end())
		throw key_error(&block);
	return *i;
}

bool BlockAllocator::has_block(const Block &block) const
{
	return find_if(blocks.begin(), blocks.end(), BlockMatch(block))!=blocks.end();
}

bool BlockAllocator::is_block_current(const Block &block) const
{
	BlockList::const_iterator end = cur_blocks_end;
	return find_if(blocks.begin(), end, BlockMatch(block))!=cur_blocks_end;
}

void BlockAllocator::reserve_more()
{
	if(blocks.empty())
		throw logic_error("no blocks");

	if(train.get_layout().get_driver().is_halted())
		return;

	BlockIter start = blocks.back();
	if(&*start==stop_at_block)
		return;
	else if(&*start==pending_block)
	{
		TrackIter track = start.track_iter();
		if(track->is_path_changing() || !track.endpoint().has_path(track->get_active_path()))
			return;
	}

	pending_block = 0;

	// See how many sensor blocks and how much track we already have
	unsigned nsens = 0;
	float dist = 0;
	for(BlockList::const_iterator i=cur_blocks_end; i!=blocks.end(); ++i)
	{
		if((*i)->get_sensor_address())
			++nsens;
		if(nsens>0)
			dist += (*i)->get_path_length(i->entry());
	}

	float approach_margin = 50*train.get_layout().get_catalogue().get_scale();
	float min_dist = train.get_controller().get_braking_distance()*1.3+approach_margin*2;

	BlockIter block = start;

	SetFlag setf(reserving);

	while(1)
	{
		BlockIter prev = block;
		block = block.next();
		if(!block || block->get_endpoints().size()<2)
			// The track ends here
			break;

		if(block->get_turnout_address() && !prev->get_turnout_address())
		{
			/* We are arriving at a turnout.  See if we have enough blocks and
			distance reserved. */
			if(nsens>=3 && dist>=min_dist)
				break;
		}

		if(!reserve_block(block))
		{
			pending_block = &*block;
			break;
		}

		if(cur_blocks_end==blocks.end())
			--cur_blocks_end;

		TrackIter track = block.track_iter();
		if(track->is_path_changing())
		{
			pending_block = &*block;
			break;
		}
		else
		{
			unsigned path = track->get_active_path();
			if(!track.endpoint().has_path(path))
			{
				path = track->get_type().coerce_path(track.entry(), path);
				track->set_active_path(path);
				if(track->is_path_changing())
				{
					pending_block = &*block;
					break;
				}
			}
		}

		if(&*block==stop_at_block)
			break;

		if(block->get_sensor_address())
			++nsens;
		if(nsens>0)
			dist += block->get_path_length(block.entry());
	}

	if(!next_sensor)
	{
		update_next_sensor(0);
		// Immediately advance to just before the next sensor
		advance_front(next_sensor);
	}
}

bool BlockAllocator::reserve_block(const BlockIter &block)
{
	/* Add it to the list first to present a consistent state in block_reserved
	signal. */
	blocks.push_back(block);

	bool first_reserve = (cur_blocks_end==blocks.end());
	if(first_reserve)
		--cur_blocks_end;

	try
	{
		if(!block->reserve(&train))
		{
			if(first_reserve)
				cur_blocks_end = blocks.end();
			blocks.pop_back();
			return false;
		}

		return true;
	}
	catch(...)
	{
		if(first_reserve)
			cur_blocks_end = blocks.end();
		blocks.pop_back();
		throw;
	}
}

void BlockAllocator::advance_front(const Block *block, bool inclusive)
{
	BlockList::iterator end;
	if(block)
	{
		end = cur_blocks_end;
		if(inclusive)
			--end;
		end = find_if(end, blocks.end(), BlockMatch(*block));
		if(inclusive && end!=blocks.end())
			++end;
	}
	else
		end = blocks.end();

	if(end==blocks.end() && blocks.back().block()==pending_block)
		--end;

	SetFlag setf(advancing);
	BlockList::iterator i = cur_blocks_end;
	// Update cur_blocks_end first to keep things consistent.
	cur_blocks_end = end;
	for(; i!=end; ++i)
		signal_advanced.emit(**i, (*i)->get_sensor());
}

void BlockAllocator::advance_front(const Sensor *sensor)
{
	if(sensor)
		advance_front(sensor->get_block(), dynamic_cast<const BeamGate *>(sensor));
	else
		advance_front(0, false);
}

void BlockAllocator::advance_back()
{
	bool rev = train.get_controller().get_reverse();
	const Vehicle &veh = train.get_vehicle(rev ? 0 : train.get_n_vehicles()-1);
	const Block &veh_block = veh.get_placement().get_position(rev ? VehiclePlacement::FRONT_AXLE : VehiclePlacement::BACK_AXLE)->get_block();

	bool pending_cleared = false;

	/* Sensors aren't guaranteed to be detriggered in order.  Go through the
	block list and locate the first sensor that's still active. */
	BlockList::iterator end = blocks.end();
	for(BlockList::iterator i=blocks.begin(); i!=cur_blocks_end; ++i)
	{
		Block *block = &**i;
		list<Sensor *> sensors;

		/* Collect all sensors from the block in the order they are expected to
		detrigger. */
		for(TrackIter j=i->track_iter(); (j && &j->get_block()==block); j=j.next())
			if(!j->get_attachments().empty())
			{
				Track::AttachmentList attachments = j->get_attachments_ordered(j.entry());
				for(Track::AttachmentList::const_iterator k=attachments.begin(); k!=attachments.end(); ++k)
					if(BeamGate *gate = dynamic_cast<BeamGate *>(*k))
						sensors.push_back(gate);
			}

		if(Sensor *sensor = (*i)->get_sensor())
			sensors.push_back(sensor);

		/* See if any sensor is still active, and record the position of the
		last inactive sensor. */
		bool active_sensor = false;
		for(list<Sensor *>::const_iterator j=sensors.begin(); (!active_sensor && j!=sensors.end()); ++j)
		{
			if((*j)->get_state())
				active_sensor = true;
			else
				end = i;
		}

		if(block==pending_block)
			pending_cleared = true;

		// Stop if we encounter an active sensor or the train's last vehicle
		if(block==&veh_block || active_sensor)
		{
			if(end!=blocks.end())
			{
				/* If the last inactive sensor was in an earlier block, release
				that block as well. */
				if(i!=end)
					++end;
				release_blocks_begin(end);
				signal_rear_advanced.emit(*last());
			}

			if(pending_cleared)
				reserve_more();

			return;
		}
	}
}

void BlockAllocator::release_blocks_begin(const BlockList::iterator &end)
{
	for(BlockList::iterator i=blocks.begin(); i!=end; )
		release_block(i++);
}

void BlockAllocator::release_blocks_end(const BlockList::iterator &begin)
{
	// Guard against decrementing blocks.begin()
	if(begin==blocks.begin())
		return release_blocks_begin(blocks.end());

	if(begin==blocks.end())
		return;

	/* Release the blocks in reverse order so that a consistent state is
	presented in block_reserved signal. */
	bool done = false;
	for(BlockList::iterator i=--blocks.end(); !done; )
	{
		done = (i==begin);
		release_block(i--);
	}
}

void BlockAllocator::release_block(const BlockList::iterator &i)
{
	if(advancing)
		throw logic_error("cannot release while advancing");
	if(i==cur_blocks_end)
		++cur_blocks_end;
	if(next_sensor && &**i==next_sensor->get_block())
		next_sensor = 0;
	if(&**i==pending_block)
		pending_block = 0;

	Block &block = **i;
	blocks.erase(i);
	block.reserve(0);
}

void BlockAllocator::reverse()
{
	release_blocks_end(cur_blocks_end);
	blocks.reverse();
	for(BlockList::iterator i=blocks.begin(); i!=blocks.end(); ++i)
		*i = i->reverse();

	if(active)
		reserve_more();
}

void BlockAllocator::turnout_path_changing(Track &track)
{
	BlockList::iterator i = find_if(cur_blocks_end, blocks.end(), BlockMatch(track.get_block()));
	if(i!=blocks.end())
	{
		++i;
		release_blocks_end(i);
		pending_block = &track.get_block();
	}
}

void BlockAllocator::turnout_path_changed(Track &track)
{
	if(&track.get_block()==pending_block && !reserving)
		reserve_more();
}

void BlockAllocator::block_reserved(Block &block, const Train *tr)
{
	if(&block==pending_block && !tr && !reserving)
		reserve_more();
}

void BlockAllocator::sensor_state_changed(Sensor &sensor, Sensor::State state)
{
	Block *block = sensor.get_block();
	if(!block || block->get_train()!=&train)
		return;

	if(state==Sensor::MAYBE_ACTIVE)
	{
		if(&sensor==next_sensor)
		{
			if(is_block_current(*block))
				signal_advanced.emit(*block, &sensor);
			update_next_sensor(next_sensor);
			advance_front(next_sensor);

			if(active)
				reserve_more();
		}
		else if(!is_block_current(*block))
			train.get_layout().emergency(block, "Sensor for "+train.get_name()+" triggered out of order");
	}
	else if(state==Sensor::INACTIVE)
		advance_back();
}

void BlockAllocator::update_next_sensor(Sensor *after)
{
	BeamGate *after_gate = dynamic_cast<BeamGate *>(after);

	BlockList::iterator i = cur_blocks_end;
	if(after)
	{
		if(after_gate)
			--i;
		i = find_if(i, blocks.end(), BlockMatch(*after->get_block()));
	}

	for(; i!=blocks.end(); ++i)
	{
		if(Sensor *sensor = (*i)->get_sensor())
		{
			if(!after_gate && sensor!=next_sensor)
			{
				next_sensor = sensor;
				return;
			}
		}

		Block *block = &**i;
		for(TrackIter j=i->track_iter(); (j && &j->get_block()==block); j=j.next())
			if(!j->get_attachments().empty())
			{
				Track::AttachmentList attachments = j->get_attachments_ordered(j.entry());
				for(Track::AttachmentList::const_iterator k=attachments.begin(); k!=attachments.end(); ++k)
					if(BeamGate *gate = dynamic_cast<BeamGate *>(*k))
					{
						if(after_gate)
						{
							if(gate==after_gate)
								after_gate = 0;
						}
						else
						{
							next_sensor = gate;
							return;
						}
					}
			}
	}

	next_sensor = 0;
}

void BlockAllocator::halt_event(bool halted)
{
	if(active && !halted)
		reserve_more();
}

void BlockAllocator::save(list<DataFile::Statement> &st) const
{
	if(!blocks.empty() && cur_blocks_end!=blocks.begin())
	{
		BlockList cur_blocks(blocks.begin(), BlockList::const_iterator(cur_blocks_end));
		BlockIter prev;
		if(train.get_controller().get_reverse())
		{
			cur_blocks.reverse();
			prev = cur_blocks.front().next();
		}
		else
			prev = cur_blocks.front().flip();

		st.push_back((DataFile::Statement("hint"), prev->get_id()));

		for(BlockList::const_iterator i=cur_blocks.begin(); (i!=cur_blocks.end() && i->block()!=pending_block); ++i)
			st.push_back((DataFile::Statement("block"), (*i)->get_id()));
	}
}


BlockAllocator::Loader::Loader(BlockAllocator &ba):
	DataFile::ObjectLoader<BlockAllocator>(ba),
	valid(true)
{
	add("block", &Loader::block);
	add("hint",  &Loader::hint);
}

void BlockAllocator::Loader::block(unsigned id)
{
	if(!valid)
		return;

	Block *blk;
	try
	{
		blk = &obj.train.get_layout().get_block(id);
	}
	catch(const key_error &)
	{
		valid = false;
		return;
	}

	int entry = -1;
	if(prev_block)
		entry = blk->get_endpoint_by_link(*prev_block);
	if(entry<0)
		entry = 0;

	obj.blocks.push_back(BlockIter(blk, entry));
	blk->reserve(&obj.train);

	if(blk->get_sensor_address())
		obj.train.get_layout().get_driver().set_sensor(blk->get_sensor_address(), true);

	prev_block = blk;
}

void BlockAllocator::Loader::hint(unsigned id)
{
	try
	{
		prev_block = &obj.train.get_layout().get_block(id);
	}
	catch(const key_error &)
	{
		valid = false;
	}
}

} // namespace R2C2
