#include "blockiter.h"
#include "catalogue.h"
#include "driver.h"
#include "layout.h"
#include "signal.h"
#include "signaltype.h"
#include "trackiter.h"
#include "tracktype.h"
#include "train.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Signal::Signal(Layout &l, const SignalType &t):
	TrackAttachment(l),
	type(t),
	address(0),
	train(0),
	check_allocated_blocks(false),
	passing(false)
{
	layout.add(*this);

	layout.signal_block_reserved.connect(sigc::mem_fun(this, &Signal::block_reserved));
}

Signal::~Signal()
{
	if(layout.has_driver() && address)
		layout.get_driver().remove_sensor(address);
	layout.remove(*this);
}

Signal *Signal::clone(Layout *to_layout) const
{
	Signal *sig = new Signal((to_layout ? *to_layout : layout), type);
	sig->set_position(position);
	sig->set_rotation(rotation);
	return sig;
}

void Signal::set_address(unsigned a)
{
	Driver *driver = (layout.has_driver() ? &layout.get_driver() : 0);

	if(driver && address)
		driver->remove_signal(address);
	address = a;
	if(driver && address)
		driver->add_signal(address, type);
}

void Signal::set_position(const Vector &p)
{
	position = p;

	update_attachment();
	signal_moved.emit();
}

void Signal::set_rotation(const Angle &r)
{
	rotation = r;

	update_attachment();
	signal_moved.emit();
}

void Signal::update_attachment()
{
	attach_to_closest(2);

	if(track)
		block = track.block_iter();
	else
		block = BlockIter();
}

unsigned Signal::get_n_snap_nodes() const
{
	return 1;
}

Snap Signal::get_snap_node(unsigned i) const
{
	if(i>=1)
		throw out_of_range("Signal::get_snap_node");

	Snap sn;
	sn.position = position;
	sn.rotation = rotation;
	return sn;
}

SnapType Signal::get_default_snap_type_to(const Object &other) const
{
	if(dynamic_cast<const Track *>(&other))
		return SNAP_SEGMENT;

	return NO_SNAP;
}

void Signal::tick(const Time::TimeDelta &)
{
	if(check_allocated_blocks)
	{
		unsigned n_blocks = 0;
		const Block *last = train->get_block_allocator().last().block();
		if(block.block()!=last)
		{
			BlockIter iter = block.next();
			while(iter && iter->get_train()==train && iter.block()!=last)
			{
				if(iter->get_sensor_address())
					++n_blocks;
				iter = iter.next();
			}
		}
		check_allocated_blocks = false;

		const list<SignalType::Indication> &indications = type.get_indications();
		unsigned aspect = indications.back().aspect;
		for(list<SignalType::Indication>::const_iterator i=indications.begin(); i!=indications.end(); ++i)
			if(n_blocks>=i->free_blocks)
			{
				aspect = i->aspect;
				break;
			}

		layout.get_driver().set_signal(address, aspect);
	}
}

void Signal::block_reserved(const Block &b, Train *t)
{
	if(&b==block.block())
	{
		if(t)
		{
			const BlockIter &b_iter = t->get_block_allocator().iter_for(b);
			if(b_iter && b_iter.entry()==block.entry())
			{
				if(train_conn)
					train_conn.disconnect();
				train = t;
				passing = false;
				train_conn = train->signal_advanced.connect(sigc::mem_fun(this, &Signal::train_advanced));
				check_allocated_blocks = true;
			}
		}
		else
		{
			layout.get_driver().set_signal(address, type.get_indications().back().aspect);
			reset();
		}
	}
	else if(train && t==train)
		check_allocated_blocks = true;
}

void Signal::train_advanced(Block &b)
{
	if(&b==block.block())
		passing = true;
	else if(passing && b.get_sensor_address())
	{
		layout.get_driver().set_signal(address, type.get_indications().back().aspect);
		reset();
	}
}

void Signal::reset()
{
	train = 0;
	if(train_conn)
		train_conn.disconnect();
	check_allocated_blocks = false;
}

void Signal::save(list<DataFile::Statement> &st) const
{
	st.push_back((DataFile::Statement("position"), position.x, position.y, position.z));
	st.push_back((DataFile::Statement("rotation"), rotation.radians()));
	if(address)
		st.push_back((DataFile::Statement("address"), address));
}


Signal::Loader::Loader(Signal &s):
	DataFile::ObjectLoader<Signal>(s)
{
	add("address",  &Loader::address);
	add("position", &Loader::position);
	add("rotation", &Loader::rotation);
}

void Signal::Loader::address(unsigned a)
{
	obj.set_address(a);
}

void Signal::Loader::position(float x, float y, float z)
{
	obj.set_position(Vector(x, y, z));
}

void Signal::Loader::rotation(float d)
{
	obj.set_rotation(Angle::from_radians(d));
}

} // namespace R2C2
