#ifndef LIBR2C2_TRACK_H_
#define LIBR2C2_TRACK_H_

#include <list>
#include <set>
#include <sigc++/trackable.h>
#include <msp/datafile/objectloader.h>
#include "geometry.h"
#include "object.h"
#include "tracktype.h"

namespace R2C2 {

class Block;
class Layout;
class TrackAttachment;

class Track: public Object, public sigc::trackable
{
public:
	class Loader: public Msp::DataFile::ObjectLoader<Track>
	{
	public:
		Loader(Track &);
	private:
		void path(unsigned);
		void position(float, float, float);
		void rotation(float);
		void sensor_address(unsigned);
		void slope(float);
		void tilt(float);
		void turnout_address(unsigned);
	};

	typedef std::list<TrackAttachment *> AttachmentList;

	sigc::signal<void, unsigned, Track *> signal_link_changed;
	sigc::signal<void, unsigned> signal_path_changing;
	sigc::signal<void, unsigned> signal_path_changed;

private:
	const TrackType &type;
	Block *block;
	float slope;
	bool flex;
	unsigned turnout_addr;
	unsigned turnout_id;
	unsigned sensor_addr;
	std::vector<Track *> links;
	unsigned active_path;
	bool path_changing;
	AttachmentList attachments;
	int preferred_exit;

	Track(const Track &);
	Track &operator=(const Track &);
public:
	Track(Layout &, const TrackType &);
	~Track();

	virtual Track *clone(Layout * = 0) const;
	virtual const TrackType &get_type() const { return type; }

	void set_block(Block *);
	Block &get_block() const;
	virtual void set_position(const Vector &);
	virtual void set_rotation(const Angle &);
	virtual void set_tilt(const Angle &);
	void set_flex(bool);
	bool get_flex() const { return flex; }
private:
	void propagate_slope();
	void check_slope();

public:
	void set_turnout_address(unsigned);
	void set_sensor_address(unsigned);
	unsigned get_turnout_address() const { return turnout_addr; }
	unsigned get_sensor_address() const { return sensor_addr; }
	void set_preferred_exit(int);
	int get_preferred_exit() const { return preferred_exit; }
	void set_active_path(unsigned);
	unsigned get_active_path() const { return active_path; }
	bool is_path_changing() const { return path_changing; }
	float get_path_length(int = -1) const;

	OrientedPoint get_point(unsigned, unsigned, float) const;
	OrientedPoint get_point(unsigned, float) const;

	virtual unsigned get_n_snap_nodes() const;
	virtual Snap get_snap_node(unsigned) const;
	virtual bool snap(Snap &, float, SnapType = SNAP_DEFAULT) const;
private:
	virtual SnapType get_default_snap_type_to(const Object &) const;

public:
	virtual unsigned get_n_link_slots() const;
	virtual Track *get_link(unsigned) const;
	const std::vector<Track *> &get_links() const { return links; }
	virtual int get_link_slot(const Object &) const;
	virtual bool link_to(Object &);
	using Object::break_link;
	virtual bool break_link(unsigned);

	void add_attachment(TrackAttachment &);
	void remove_attachment(TrackAttachment &);
	const AttachmentList &get_attachments() const { return attachments; }
	AttachmentList get_attachments_ordered(unsigned) const;

	void save(std::list<Msp::DataFile::Statement> &) const;
	void save_dynamic(std::list<Msp::DataFile::Statement> &) const;
private:
	void turnout_event(unsigned, unsigned);
	void turnout_failed(unsigned);
};

} // namespace R2C2

#endif
